from enum import Enum

class StepType(Enum):
    START = 1
    END = 2

class PartialLogEntry:
    def __init__(self):
        self.name = ""
        self.id = ""
        self.type = StepType.START
        self.time = 0
        self.display_time = ""
        self.instance_id = ""
        



from dataclasses import dataclass

@dataclass(init=False)
class PerformanceLogEntry:
    name: str
    id: str
    instance_id: str
    start_time: int
    end_time: int
    start_display_time: str
    duration: int
    is_test_passed: bool
    instance_id: str

    def __init__(self):
        self.name = ""
        self.id = ""
        self.instance_id = ""
        self.start_time = 0
        self.end_time = 0
        self.start_display_time = ""
        self.duration = 0
        self.is_test_passed = True
        self.instance_id = ""


    def get_duration(self) -> int:
        return self.end_time - self.start_time


import math


class Calculator:
    def get_average_and_standard_deviation(self, duration_list):
        mean = self.get_average_time_span(duration_list)
        sos = 0
        sem = 0

        # calculate sem
        if len(duration_list) > 1:
            for d in duration_list:
                sos += math.pow(d - mean, 2)
            std = math.sqrt(sos / (len(duration_list) - 1))
            sem = std / math.sqrt(len(duration_list))
        return [round(mean), round(sem)]

    def get_average_time_span(self, duration_list):
        result = sum(duration_list) / len(duration_list)
        return round(result)



from helpers.bcolors import Bcolors


def print_dict_as_table(dict_list):
    if not dict_list or not all(isinstance(d, dict) for d in dict_list):
        return
    else:
        print("\npytest_performancetotal results:")
    headers = dict_list[0].keys()
    column_widths = {key: max(len(str(key)), max(len(str(d.get(key, ''))) for d in dict_list)) for key in headers}
    header_row = " | ".join(f"{key:{column_widths[key]}}" for key in headers)
    separator = "-+-".join('-' * column_widths[key] for key in headers)
    print(f"{Bcolors.HEADER} {header_row}")
    print(separator)
    for d in dict_list:
        row = " | ".join(f"{str(d.get(key, '')):{column_widths[key]}}" for key in headers)
        print(f"{Bcolors.OKCYAN} {row} {Bcolors.ENDC}")


from itertools import groupby
from operator import itemgetter

class Group:
    def group_by(self, dict_list, attributes):
        dict_list.sort(key=itemgetter(*attributes))
        grouped_data = {key: list(group) for key, group in groupby(dict_list, key=itemgetter(*attributes))}
        return grouped_data

# Example usage:
# Group by 'type' and 'color'
# grouped_items = group_by(items, ['type', 'color'])



import csv
import json
from helpers.bcolors import Bcolors
from helpers.calculator import Calculator
from helpers.dic_builder import DictBuilder
from helpers.file_writer import FileWriter
from helpers.group import Group
from helpers.logger import Logger
from helpers.table_printer import print_dict_as_table

class PerformanceAnalyzer:
    def __init__(self):
        self._performance_results = []
        self.logger = Logger()
    

    def analyze(
        self, log_file_name, save_data_file_path, drop_results_from_failed_test
    ):
        performance_log_entries = self.deserialize_data(log_file_name)
        if drop_results_from_failed_test:
            performance_log_entries = [
                e for e in performance_log_entries if e["is_test_passed"]
            ]
        if not performance_log_entries:
            return
        grouped_results = Group().group_by(performance_log_entries, ["name"])
        for group_key, group_items in grouped_results.items():
            duration_list = [item["duration"] for item in group_items]
            avg_and_ste = Calculator().get_average_and_standard_deviation(duration_list)
            performance_result = (
                DictBuilder()
                .add("name", group_items[0]["name"])
                .add("earliest_time", group_items[0]["start_display_time"])
                .add("latest_time", group_items[-1]["start_display_time"])
                .add("avg_time", avg_and_ste[0])
                .add("sem", avg_and_ste[1])
                .add("repeats", len(duration_list))
                .add("min_value", min(duration_list))
                .add("max_value", max(duration_list))
                .build()
            )
            self._performance_results.append(performance_result)
        self.serialize_data(save_data_file_path)

    def serialize_data(self, save_data_file_path):
        # Print to the cli
        printed_keys = {'name', 'avg_time', 'sem', 'repeats', 'min_value', 'max_value'}
        printed_data = [{k: d[k] for k in printed_keys if k in d} for d in self._performance_results]
        print_dict_as_table(printed_data)
        
        # Serialize to JSON
        file_writer = FileWriter()
        file_writer.write_to_file(
            f"{save_data_file_path}.json", json.dumps(self._performance_results)
        )

        # Serialize to CSV
        try:
            with open(f"{save_data_file_path}.csv", mode="w", newline="") as csv_file:
                fieldnames = [
                    "name",
                    "avg_time",
                    "sem",
                    "repeats",
                    "min_value",
                    "max_value",
                    "earliest_time",
                    "latest_time"
                ]

                csv_writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
                csv_writer.writeheader()
                for result in self._performance_results:
                    row = {
                        "name": result["name"],
                        "avg_time": result["avg_time"],
                        "sem": result["sem"],
                        "repeats": result["repeats"],
                        "min_value": result["min_value"],
                        "max_value": result["max_value"],
                        "earliest_time":result["earliest_time"],
                        "latest_time":result["latest_time"]
                    }
                    csv_writer.writerow(row)
                self.logger.info(f"{Bcolors.OKCYAN}pytest_performancetotal: results saved to: {save_data_file_path} {Bcolors.ENDC}")
        except IOError as e:
            self.logger.error(f"{Bcolors.WARNING}pytest_performancetotal: error occurred while writing to the CSV file: {e}. Make sure the file is not open in another program. {Bcolors.ENDC}")

    def deserialize_data(self, file_name: str):
        results_array = []
        try:
            text_results_array = FileWriter().read_all_lines(file_name)
            for text_result in text_results_array:
                if text_result != "":
                    performance_result = json.loads(text_result)

                    if "id" in performance_result:
                        results_array.append(performance_result)
        except Exception as e:
            self.logger.error(f"{Bcolors.WARNING}pytest_performancetotal: error occurred while reading file {file_name}: {e} {Bcolors.ENDC}")
        return results_array



from datetime import datetime
import json
from entities.partial_log_entry import PartialLogEntry, StepType
from entities.performance_log_entry import PerformanceLogEntry
from helpers.file_writer import FileWriter
from helpers.id_generator import IdGenerator
from consts.constants import constant_vars


class PerformanceCache:
    def __init__(self):
        self._start_log_entries = []
        self._end_log_entries = []
        self._performance_entries = []

    def sample_start(self, step_name, instance_id):
        log_entry = self.set_sample(StepType.START, step_name, instance_id)
        self._start_log_entries.insert(0, log_entry)

    def sample_end(self, step_name, instance_id):
        log_entry = self.set_sample(StepType.END, step_name, instance_id)
        self._end_log_entries.append(log_entry)

    def get_sample_time(self, step_name):
        return self.get_performance_entry_time(step_name)

    def flush(self, file_name, is_test_passed):
        self.create_performance_entries(is_test_passed)
        self.write_performance_data_to_file(file_name)

    def set_sample(self, step_type, step_name, instance_id):
        id = ""
        log_entry = PartialLogEntry()

        if step_type == StepType.START:
            id = IdGenerator().get_id()
        else:
            id = self.get_start_id_by_step_name(step_name, instance_id)

        log_entry.id = id
        log_entry.instance_id = instance_id
        log_entry.name = step_name
        log_entry.type = step_type
        log_entry.time = int(datetime.now().timestamp() * 1000)
        log_entry.display_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        return log_entry

    def get_performance_entry_time(self, step_name):
        duration = 0
        start_entry = next(
            (
                e
                for e in self._start_log_entries
                if e.name == step_name + constant_vars["step_suffix_used"]
            ),
            None,
        )
        if start_entry:
            end_entry = next(
                (e for e in self._end_log_entries if e.id == start_entry.id), None
            )
            if end_entry:
                duration = end_entry.time - start_entry.time
        return duration

    def create_performance_entries(self, is_test_passed):
        rev_start_entries = list(reversed(self._start_log_entries))
        for start_entry in rev_start_entries:
            temp_performance_entry = PerformanceLogEntry()
            corresponded_end_entry = next(
                (e for e in self._end_log_entries if e.id == start_entry.id), None
            )
            if corresponded_end_entry:
                temp_performance_entry.id = start_entry.id
                temp_performance_entry.instance_id = start_entry.instance_id
                temp_performance_entry.name = corresponded_end_entry.name
                temp_performance_entry.start_display_time = start_entry.display_time
                temp_performance_entry.start_time = start_entry.time
                temp_performance_entry.end_time = corresponded_end_entry.time
                temp_performance_entry.duration = temp_performance_entry.get_duration()
                temp_performance_entry.is_test_passed = is_test_passed
                self._performance_entries.append(temp_performance_entry)
    def get_start_id_by_step_name(self, step_name: str, instance_id: str) -> str:
        id = ""
        start_entry = next(
            (
                e
                for e in self._start_log_entries
                if e.name == step_name and e.instance_id == instance_id
            ),
            None,
        )
        if start_entry:
            id = start_entry.id
        else:
            return
        start_entry.name += constant_vars.get("step_suffix_used")
        return id

    def clear_data(self) -> None:
        self._start_log_entries = []
        self._end_log_entries = []
        self._performance_entries = []

    def write_performance_data_to_file(self, file_name: str) -> None:
        for performance_entry in self._performance_entries:
            FileWriter().append_line_to_file(
                file_name, f"{json.dumps(performance_entry.__dict__)}\n"
            )
        self.clear_data()



import os
from typing import List
from filelock import FileLock
from helpers.bcolors import Bcolors
from helpers.logger import Logger


class FileWriter:
    def __init__(self):
        self.logger = Logger()

    def get_file_lock(self, data_file_path):
        lock_file_path = f"{data_file_path}.lock"
        return FileLock(lock_file_path)

    def read_all_lines(self, path: str) -> List[str]:
        lock = self.get_file_lock(path)
        try:
            with lock:
                with open(path, "r") as file:
                    data = file.read()
        except OSError as err:
            self.logger.error(f"{Bcolors.WARNING}pytest_performancetotal: failed to read file {path}: {err} {Bcolors.ENDC}")
            return []
        return data.splitlines()

    def write_to_file(self, path: str, data: str) -> None:
        lock = self.get_file_lock(path)
        try:
            with lock:
                with open(path, "w") as file:
                    file.write(data)
        except OSError as err:
            self.logger.error(
                f"{Bcolors.WARNING}pytest_performancetotal error: '{path}' is not writable: {err} {Bcolors.ENDC}"
            )

    def append_line_to_file(self, path: str, data: str) -> None:
        lock = self.get_file_lock(path)
        try:
            with lock:
                with open(path, "a") as file:
                    file.write(data + "\n")
        except OSError as err:
            self.logger.error(
                f"{Bcolors.WARNING}pytest_performancetotal error: file '{path}' is not writable (append): {err} {Bcolors.ENDC}"
            )

    def get_file_path(self, results_dir: str, file_name: str) -> str:
        path = os.path.join(results_dir, file_name)
        return path

    def create_results_dir_if_not_exist(self, root_dir, results_dirname) -> str:
        dir_path = os.path.join(str(root_dir), results_dirname)
        is_file_exists = self.is_file_exist(dir_path)
        if not is_file_exists:
            self.make_dir(dir_path)
        return dir_path

    def is_file_exist(self, dir_path: str) -> bool:
        return os.path.exists(dir_path)

    def make_dir(self, dir_path: str) -> None:
        try:
            os.mkdir(dir_path)
        except OSError as err:
            self.logger.error(
                f"{Bcolors.WARNING}pytest_performancetotal: can't create directory '{dir_path}': {err} {Bcolors.ENDC}"
            )



from datetime import datetime
import json
import os
from pytest import FixtureRequest
from pytest_performancetotal.performance_analyzer import PerformanceAnalyzer
from helpers.file_writer import FileWriter
from helpers.id_generator import IdGenerator
from pytest_performancetotal.performance_cache import PerformanceCache


class Performance:

    def __init__(self):
        self.instance_id = IdGenerator().get_id("inst")
        self.performance_cache = PerformanceCache()
        self.performance_results_directory_name = "performance_results"
        self.log_file_name = "performance-log.txt"

    def sample_start(self, stepName: str) -> None:
        self.performance_cache.sample_start(stepName, self.instance_id)

    def sample_end(self, stepName: str) -> None:
        self.performance_cache.sample_end(stepName, self.instance_id)

    def get_sample_time(self, stepName: str) -> int:
        return self.performance_cache.get_sample_time(stepName)

    def initialize(self, root_dir, no_append) -> None:
        """Don't use this method directly."""
        global performancetotal_results_dir_path
        resultsDirPath = ""
        file_writer = FileWriter()
        if "performancetotal_results_dir_path" in globals():
            resultsDirPath = performancetotal_results_dir_path
        else:
            resultsDirPath = file_writer.create_results_dir_if_not_exist(
                root_dir, self.performance_results_directory_name
            )
            performancetotal_results_dir_path = resultsDirPath
        local_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        init_obj = json.dumps(
            {"start_display_time": local_time, "instance_id": self.instance_id}
        )
        file_name = os.path.join(resultsDirPath, self.log_file_name)
        if no_append:
            file_writer.write_to_file(file_name, f"{init_obj}\n")
        else:
            file_writer.append_line_to_file(file_name, f"{init_obj}\n")

    def finalize_test(self, request: FixtureRequest) -> None:
        """Don't use this method directly."""
        global performancetotal_results_dir_path
        test_passed = hasattr(request.node, "rep_call") and request.node.rep_call.passed
        self.performance_cache.flush(
            FileWriter().get_file_path(
                performancetotal_results_dir_path, self.log_file_name
            ),
            test_passed,
        )

    def analyze_results(self, workerIndex: int) -> None:
        """Don't use this method directly."""
        analyzer = PerformanceAnalyzer()
        file_writer = FileWriter()
        results_dir = performancetotal_results_dir_path
        log_file_name = file_writer.get_file_path(results_dir, self.log_file_name)
        save_data_file_path = file_writer.get_file_path(results_dir, "results")
        analyzer.analyze(log_file_name, save_data_file_path, False)



from setuptools import find_packages, setup

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='pytest-performancetotal',
    packages=find_packages(where='src', exclude=['*tests*']),
    package_dir={'': 'src'},
    version='0.2.4',
    author='Tzur Paldi',
    author_email='tzur.paldi@outlook.com',
    maintainer='Tzur Paldi',
    maintainer_email='tzur.paldi@outlook.com',
    license='MIT',
    url='https://github.com/tzurp/pytest_performancetotal',
    description='A performance plugin for pytest',
    keywords='pytest plugin performance playwright',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Pytest",
    ],
    entry_points={
        'pytest11': [
            'pytest_performancetotal = pytest_performancetotal.performancetotal',
        ],
    },
    install_requires=[
        'filelock>=3.13.0',
    ]
)



from typing import Generator
import pytest
from pytest_performancetotal.performance import Performance

def pytest_addoption(parser: pytest.Parser):
    parser.addoption("--performance-noappend", action='store_true', help='Disables results appending to existing file. Previous data is deleted.')
    
@pytest.fixture
def performancetotal(request) -> Generator[Performance, None, None]:
    performance = Performance()
    yield performance
    performance.finalize_test(request)

@pytest.fixture(scope='session', autouse=True)
def performance_worker(request: pytest.FixtureRequest) -> Generator[None, None, None]:
    performance = Performance()
    no_append = request.config.getoption('--performance-noappend', False)
    performance.initialize(request.config.rootdir, no_append)
    yield
    session_id = request.session.nodeid
    performance.analyze_results(session_id)
    
    